import { expect } from "chai";
import { HdKeyType, Wif } from "../lib";
import { HdPrivateKey } from "../lib/HdPrivateKey";
import { HdPublicKey } from "../lib/HdPublicKey";
import { Network } from "../lib/Network";

describe("HdKey", () => {
    function assertExtPrivateKey(result: HdPrivateKey, prv: string) {
        const exp = HdPrivateKey.decode(prv);
        expect(result.version).to.equal(exp.version);
        expect(result.depth).to.equal(exp.depth);
        expect(result.number).to.equal(exp.number);
        expect(result.privateKey.toHex()).to.equal(exp.privateKey.toHex());
        expect(result.chainCode.toString("hex")).to.equal(exp.chainCode.toString("hex"));
        expect(result.parentFingerprint.toString("hex")).to.equal(
            exp.parentFingerprint.toString("hex"),
        );
    }

    function assertExtPublicKey(result: HdPublicKey, pub: string) {
        const exp = HdPublicKey.decode(pub);
        expect(result.version).to.equal(exp.version);
        expect(result.depth).to.equal(exp.depth);
        expect(result.number).to.equal(exp.number);
        expect(result.publicKey.toHex()).to.equal(exp.publicKey.toHex());
        expect(result.chainCode.toString("hex")).to.equal(exp.chainCode.toString("hex"));
        expect(result.parentFingerprint.toString("hex")).to.equal(
            exp.parentFingerprint.toString("hex"),
        );
    }

    function assertChain(
        seed: Buffer,
        vectors: [string, string, string][],
        network: Network,
        type: HdKeyType,
    ) {
        for (const [path, prv, pub] of vectors) {
            it(path, () => {
                const privKey = HdPrivateKey.fromPath(path, seed, network, type);

                assertExtPrivateKey(privKey, prv);
                expect(privKey.encode()).to.equal(prv);

                if (!pub) return;
                const pubKey = privKey.toPubKey();
                assertExtPublicKey(pubKey, pub);
                expect(pubKey.encode()).to.equal(pub);
            });
        }
    }

    describe("BIP32", () => {
        describe("vector 1", () => {
            const seed = Buffer.from("000102030405060708090a0b0c0d0e0f", "hex");

            const vectors: [string, string, string][] = [
                [
                    "m",
                    "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi",
                    "xpub661MyMwAqRbcFtXgS5sYJABqqG9YLmC4Q1Rdap9gSE8NqtwybGhePY2gZ29ESFjqJoCu1Rupje8YtGqsefD265TMg7usUDFdp6W1EGMcet8",
                ],
                [
                    "m/0'",
                    "xprv9uHRZZhk6KAJC1avXpDAp4MDc3sQKNxDiPvvkX8Br5ngLNv1TxvUxt4cV1rGL5hj6KCesnDYUhd7oWgT11eZG7XnxHrnYeSvkzY7d2bhkJ7",
                    "xpub68Gmy5EdvgibQVfPdqkBBCHxA5htiqg55crXYuXoQRKfDBFA1WEjWgP6LHhwBZeNK1VTsfTFUHCdrfp1bgwQ9xv5ski8PX9rL2dZXvgGDnw",
                ],
                [
                    "m/0'/1",
                    "xprv9wTYmMFdV23N2TdNG573QoEsfRrWKQgWeibmLntzniatZvR9BmLnvSxqu53Kw1UmYPxLgboyZQaXwTCg8MSY3H2EU4pWcQDnRnrVA1xe8fs",
                    "xpub6ASuArnXKPbfEwhqN6e3mwBcDTgzisQN1wXN9BJcM47sSikHjJf3UFHKkNAWbWMiGj7Wf5uMash7SyYq527Hqck2AxYysAA7xmALppuCkwQ",
                ],
                [
                    "m/0'/1/2'",
                    "xprv9z4pot5VBttmtdRTWfWQmoH1taj2axGVzFqSb8C9xaxKymcFzXBDptWmT7FwuEzG3ryjH4ktypQSAewRiNMjANTtpgP4mLTj34bhnZX7UiM",
                    "xpub6D4BDPcP2GT577Vvch3R8wDkScZWzQzMMUm3PWbmWvVJrZwQY4VUNgqFJPMM3No2dFDFGTsxxpG5uJh7n7epu4trkrX7x7DogT5Uv6fcLW5",
                ],
                [
                    "m/0'/1/2'/2",
                    "xprvA2JDeKCSNNZky6uBCviVfJSKyQ1mDYahRjijr5idH2WwLsEd4Hsb2Tyh8RfQMuPh7f7RtyzTtdrbdqqsunu5Mm3wDvUAKRHSC34sJ7in334",
                    "xpub6FHa3pjLCk84BayeJxFW2SP4XRrFd1JYnxeLeU8EqN3vDfZmbqBqaGJAyiLjTAwm6ZLRQUMv1ZACTj37sR62cfN7fe5JnJ7dh8zL4fiyLHV",
                ],
                [
                    "m/0'/1/2'/2/1000000000",
                    "xprvA41z7zogVVwxVSgdKUHDy1SKmdb533PjDz7J6N6mV6uS3ze1ai8FHa8kmHScGpWmj4WggLyQjgPie1rFSruoUihUZREPSL39UNdE3BBDu76",
                    "xpub6H1LXWLaKsWFhvm6RVpEL9P4KfRZSW7abD2ttkWP3SSQvnyA8FSVqNTEcYFgJS2UaFcxupHiYkro49S8yGasTvXEYBVPamhGW6cFJodrTHy",
                ],
            ];

            assertChain(seed, vectors, Network.mainnet, HdKeyType.x);
        });

        describe("vector 2", () => {
            const seed = Buffer.from(
                "fffcf9f6f3f0edeae7e4e1dedbd8d5d2cfccc9c6c3c0bdbab7b4b1aeaba8a5a29f9c999693908d8a8784817e7b7875726f6c696663605d5a5754514e4b484542",
                "hex",
            );

            const vectors: [string, string, string][] = [
                [
                    "m",
                    "xprv9s21ZrQH143K31xYSDQpPDxsXRTUcvj2iNHm5NUtrGiGG5e2DtALGdso3pGz6ssrdK4PFmM8NSpSBHNqPqm55Qn3LqFtT2emdEXVYsCzC2U",
                    "xpub661MyMwAqRbcFW31YEwpkMuc5THy2PSt5bDMsktWQcFF8syAmRUapSCGu8ED9W6oDMSgv6Zz8idoc4a6mr8BDzTJY47LJhkJ8UB7WEGuduB",
                ],
                [
                    "m/0",
                    "xprv9vHkqa6EV4sPZHYqZznhT2NPtPCjKuDKGY38FBWLvgaDx45zo9WQRUT3dKYnjwih2yJD9mkrocEZXo1ex8G81dwSM1fwqWpWkeS3v86pgKt",
                    "xpub69H7F5d8KSRgmmdJg2KhpAK8SR3DjMwAdkxj3ZuxV27CprR9LgpeyGmXUbC6wb7ERfvrnKZjXoUmmDznezpbZb7ap6r1D3tgFxHmwMkQTPH",
                ],
                [
                    "m/0/2147483647'",
                    "xprv9wSp6B7kry3Vj9m1zSnLvN3xH8RdsPP1Mh7fAaR7aRLcQMKTR2vidYEeEg2mUCTAwCd6vnxVrcjfy2kRgVsFawNzmjuHc2YmYRmagcEPdU9",
                    "xpub6ASAVgeehLbnwdqV6UKMHVzgqAG8Gr6riv3Fxxpj8ksbH9ebxaEyBLZ85ySDhKiLDBrQSARLq1uNRts8RuJiHjaDMBU4Zn9h8LZNnBC5y4a",
                ],
                [
                    "m/0/2147483647'/1",
                    "xprv9zFnWC6h2cLgpmSA46vutJzBcfJ8yaJGg8cX1e5StJh45BBciYTRXSd25UEPVuesF9yog62tGAQtHjXajPPdbRCHuWS6T8XA2ECKADdw4Ef",
                    "xpub6DF8uhdarytz3FWdA8TvFSvvAh8dP3283MY7p2V4SeE2wyWmG5mg5EwVvmdMVCQcoNJxGoWaU9DCWh89LojfZ537wTfunKau47EL2dhHKon",
                ],
                [
                    "m/0/2147483647'/1/2147483646'",
                    "xprvA1RpRA33e1JQ7ifknakTFpgNXPmW2YvmhqLQYMmrj4xJXXWYpDPS3xz7iAxn8L39njGVyuoseXzU6rcxFLJ8HFsTjSyQbLYnMpCqE2VbFWc",
                    "xpub6ERApfZwUNrhLCkDtcHTcxd75RbzS1ed54G1LkBUHQVHQKqhMkhgbmJbZRkrgZw4koxb5JaHWkY4ALHY2grBGRjaDMzQLcgJvLJuZZvRcEL",
                ],
                [
                    "m/0/2147483647'/1/2147483646'/2",
                    "xprvA2nrNbFZABcdryreWet9Ea4LvTJcGsqrMzxHx98MMrotbir7yrKCEXw7nadnHM8Dq38EGfSh6dqA9QWTyefMLEcBYJUuekgW4BYPJcr9E7j",
                    "xpub6FnCn6nSzZAw5Tw7cgR9bi15UV96gLZhjDstkXXxvCLsUXBGXPdSnLFbdpq8p9HmGsApME5hQTZ3emM2rnY5agb9rXpVGyy3bdW6EEgAtqt",
                ],
            ];

            assertChain(seed, vectors, Network.mainnet, HdKeyType.x);
        });

        describe("vector 3", () => {
            const seed = Buffer.from(
                "4b381541583be4423346c643850da4b320e46a87ae3d2a4e6da11eba819cd4acba45d239319ac14f863b8d5ab5a0d0c64d2e8a1e7d1457df2e5a3c51c73235be",
                "hex",
            );

            const vectors: [string, string, string][] = [
                [
                    "m",
                    "xprv9s21ZrQH143K25QhxbucbDDuQ4naNntJRi4KUfWT7xo4EKsHt2QJDu7KXp1A3u7Bi1j8ph3EGsZ9Xvz9dGuVrtHHs7pXeTzjuxBrCmmhgC6",
                    "xpub661MyMwAqRbcEZVB4dScxMAdx6d4nFc9nvyvH3v4gJL378CSRZiYmhRoP7mBy6gSPSCYk6SzXPTf3ND1cZAceL7SfJ1Z3GC8vBgp2epUt13",
                ],
                [
                    "m/0'",
                    "xprv9uPDJpEQgRQfDcW7BkF7eTya6RPxXeJCqCJGHuCJ4GiRVLzkTXBAJMu2qaMWPrS7AANYqdq6vcBcBUdJCVVFceUvJFjaPdGZ2y9WACViL4L",
                    "xpub68NZiKmJWnxxS6aaHmn81bvJeTESw724CRDs6HbuccFQN9Ku14VQrADWgqbhhTHBaohPX4CjNLf9fq9MYo6oDaPPLPxSb7gwQN3ih19Zm4Y",
                ],
            ];

            assertChain(seed, vectors, Network.mainnet, HdKeyType.x);
        });

        describe("vector 4", () => {
            const seed = Buffer.from(
                "3ddd5602285899a946114506157c7997e5444528f3003f6134712147db19b678",
                "hex",
            );

            const vectors: [string, string, string][] = [
                [
                    "m",
                    "xprv9s21ZrQH143K48vGoLGRPxgo2JNkJ3J3fqkirQC2zVdk5Dgd5w14S7fRDyHH4dWNHUgkvsvNDCkvAwcSHNAQwhwgNMgZhLtQC63zxwhQmRv",
                    "xpub661MyMwAqRbcGczjuMoRm6dXaLDEhW1u34gKenbeYqAix21mdUKJyuyu5F1rzYGVxyL6tmgBUAEPrEz92mBXjByMRiJdba9wpnN37RLLAXa",
                ],
                [
                    "m/0'",
                    "xprv9vB7xEWwNp9kh1wQRfCCQMnZUEG21LpbR9NPCNN1dwhiZkjjeGRnaALmPXCX7SgjFTiCTT6bXes17boXtjq3xLpcDjzEuGLQBM5ohqkao9G",
                    "xpub69AUMk3qDBi3uW1sXgjCmVjJ2G6WQoYSnNHyzkmdCHEhSZ4tBok37xfFEqHd2AddP56Tqp4o56AePAgCjYdvpW2PU2jbUPFKsav5ut6Ch1m",
                ],
                [
                    "m/0'/1'",
                    "xprv9xJocDuwtYCMNAo3Zw76WENQeAS6WGXQ55RCy7tDJ8oALr4FWkuVoHJeHVAcAqiZLE7Je3vZJHxspZdFHfnBEjHqU5hG1Jaj32dVoS6XLT1",
                    "xpub6BJA1jSqiukeaesWfxe6sNK9CCGaujFFSJLomWHprUL9DePQ4JDkM5d88n49sMGJxrhpjazuXYWdMf17C9T5XnxkopaeS7jGk1GyyVziaMt",
                ],
            ];

            assertChain(seed, vectors, Network.mainnet, HdKeyType.x);
        });

        describe("vector 5", () => {
            const vectors = [
                [
                    "xpub661MyMwAqRbcEYS8w7XLSVeEsBXy79zSzH1J8vCdxAZningWLdN3zgtU6LBpB85b3D2yc8sfvZU521AAwdZafEz7mnzBBsz4wKY5fTtTQBm",
                    "xpub (pubkey version / prvkey mismatch)",
                ],
                [
                    "xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzFGTQQD3dC4H2D5GBj7vWvSQaaBv5cxi9gafk7NF3pnBju6dwKvH",
                    "xprv (prvkey version / pubkey mismatch)",
                ],
                [
                    "xpub661MyMwAqRbcEYS8w7XLSVeEsBXy79zSzH1J8vCdxAZningWLdN3zgtU6Txnt3siSujt9RCVYsx4qHZGc62TG4McvMGcAUjeuwZdduYEvFn",
                    "xpub (invalid pubkey prefix 04)",
                ],
                [
                    "xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzFGpWnsj83BHtEy5Zt8CcDr1UiRXuWCmTQLxEK9vbz5gPstX92JQ",
                    "xprv (invalid prvkey prefix 04)",
                ],
                [
                    "xpub661MyMwAqRbcEYS8w7XLSVeEsBXy79zSzH1J8vCdxAZningWLdN3zgtU6N8ZMMXctdiCjxTNq964yKkwrkBJJwpzZS4HS2fxvyYUA4q2Xe4",
                    "xpub (invalid pubkey prefix 01)",
                ],
                [
                    "xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzFAzHGBP2UuGCqWLTAPLcMtD9y5gkZ6Eq3Rjuahrv17fEQ3Qen6J",
                    "xprv (invalid prvkey prefix 01)",
                ],
                [
                    "xprv9s2SPatNQ9Vc6GTbVMFPFo7jsaZySyzk7L8n2uqKXJen3KUmvQNTuLh3fhZMBoG3G4ZW1N2kZuHEPY53qmbZzCHshoQnNf4GvELZfqTUrcv",
                    "xprv (zero depth with non-zero parent fingerprint)",
                ],
                [
                    "xpub661no6RGEX3uJkY4bNnPcw4URcQTrSibUZ4NqJEw5eBkv7ovTwgiT91XX27VbEXGENhYRCf7hyEbWrR3FewATdCEebj6znwMfQkhRYHRLpJ",
                    "xpub (zero depth with non-zero parent fingerprint)",
                ],
                [
                    "xprv9s21ZrQH4r4TsiLvyLXqM9P7k1K3EYhA1kkD6xuquB5i39AU8KF42acDyL3qsDbU9NmZn6MsGSUYZEsuoePmjzsB3eFKSUEh3Gu1N3cqVUN",
                    "xprv (zero depth with non-zero index)",
                ],
                [
                    "xpub661MyMwAuDcm6CRQ5N4qiHKrJ39Xe1R1NyfouMKTTWcguwVcfrZJaNvhpebzGerh7gucBvzEQWRugZDuDXjNDRmXzSZe4c7mnTK97pTvGS8",
                    "xpub (zero depth with non-zero index)",
                ],
                [
                    "DMwo58pR1QLEFihHiXPVykYB6fJmsTeHvyTp7hRThAtCX8CvYzgPcn8XnmdfHGMQzT7ayAmfo4z3gY5KfbrZWZ6St24UVf2Qgo6oujFktLHdHY4",
                    "(unknown extended key version)",
                ],
                [
                    "DMwo58pR1QLEFihHiXPVykYB6fJmsTeHvyTp7hRThAtCX8CvYzgPcn8XnmdfHPmHJiEDXkTiJTVV9rHEBUem2mwVbbNfvT2MTcAqj3nesx8uBf9",
                    "(unknown extended key version)",
                ],
                [
                    "xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzF93Y5wvzdUayhgkkFoicQZcP3y52uPPxFnfoLZB21Teqt1VvEHx",
                    "xprv (unknown extended key version)",
                ],
                [
                    "xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzFAzHGBP2UuGCqWLTAPLcMtD5SDKr24z3aiUvKr9bJpdrcLg1y3G",
                    "xprv (private key n not in 1..n-1)",
                ],
                [
                    "xpub661MyMwAqRbcEYS8w7XLSVeEsBXy79zSzH1J8vCdxAZningWLdN3zgtU6Q5JXayek4PRsn35jii4veMimro1xefsM58PgBMrvdYre8QyULY",
                    "xpub (invalid pubkey 020000000000000000000000000000000000000000000000000000000000000007)",
                ],
                [
                    "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHL",
                    "xprv (invalid checksum)",
                ],
            ];

            for (const [vector, title] of vectors) {
                it("invalid " + title, () => {
                    expect(() => HdPublicKey.decode(vector)).to.throw();
                });
            }
        });
    });

    describe("BIP49", () => {
        // abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon about
        const seed = Buffer.from(
            "5eb00bbddcf069084889a8ab9155568165f5c453ccb85e70811aaed6f6da5fc19a5ac40b389cd370d086206dec8aa6c43daea6690f20ad3d8d48b2d2ce9e38e4",
            "hex",
        );

        const vectors: [string, string, string][] = [
            [
                "m",
                "uprv8tXDerPXZ1QsVNjUJWTurs9kA1KGfKUAts74GCkcXtU8GwnH33GDRbNJpEqTvipfCyycARtQJhmdfWf8oKt41X9LL1zeD2pLsWmxEk3VAwd",
                "",
            ],
            [
                "m/49'/1'/0'",
                "uprv91G7gZkzehuMVxDJTYE6tLivdF8e4rvzSu1LFfKw3b2Qx1Aj8vpoFnHdfUZ3hmi9jsvPifmZ24RTN2KhwB8BfMLTVqaBReibyaFFcTP1s9n",
                "upub5EFU65HtV5TeiSHmZZm7FUffBGy8UKeqp7vw43jYbvZPpoVsgU93oac7Wk3u6moKegAEWtGNF8DehrnHtv21XXEMYRUocHqguyjknFHYfgY",
            ],
        ];

        assertChain(seed, vectors, Network.testnet, HdKeyType.y);

        it("creates address", () => {
            const prvKey = HdPrivateKey.fromPath(
                "m/49'/1'/0'/0/0",
                seed,
                Network.testnet,
                HdKeyType.y,
            );

            expect(prvKey.privateKey.toHex()).to.equal("c9bdb49cfbaedca21c4b1f3a7803c34636b1d7dc55a717132443fc3f4c5867e8"); // prettier-ignore
            expect(prvKey.toPubKey().publicKey.toHex()).to.equal("03a1af804ac108a8a51782198c2d034b28bf90c8803f5a53f76276fa69a4eae77f"); // prettier-ignore

            const address = prvKey.toPubKey().toAddress();
            expect(address).to.equal("2Mww8dCYPUpKHofjgcXcBCEGmniw9CoaiD2");
        });
    });

    describe("BIP84", () => {
        // abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon about
        const seed = Buffer.from(
            "5eb00bbddcf069084889a8ab9155568165f5c453ccb85e70811aaed6f6da5fc19a5ac40b389cd370d086206dec8aa6c43daea6690f20ad3d8d48b2d2ce9e38e4",
            "hex",
        );

        const vectors: [string, string, string][] = [
            [
                "m",
                "zprvAWgYBBk7JR8Gjrh4UJQ2uJdG1r3WNRRfURiABBE3RvMXYSrRJL62XuezvGdPvG6GFBZduosCc1YP5wixPox7zhZLfiUm8aunE96BBa4Kei5",
                "zpub6jftahH18ngZxLmXaKw3GSZzZsszmt9WqedkyZdezFtWRFBZqsQH5hyUmb4pCEeZGmVfQuP5bedXTB8is6fTv19U1GQRyQUKQGUTzyHACMF",
            ],
            [
                "m/84'/0'/0'",
                "zprvAdG4iTXWBoARxkkzNpNh8r6Qag3irQB8PzEMkAFeTRXxHpbF9z4QgEvBRmfvqWvGp42t42nvgGpNgYSJA9iefm1yYNZKEm7z6qUWCroSQnE",
                "zpub6rFR7y4Q2AijBEqTUquhVz398htDFrtymD9xYYfG1m4wAcvPhXNfE3EfH1r1ADqtfSdVCToUG868RvUUkgDKf31mGDtKsAYz2oz2AGutZYs",
            ],
        ];

        assertChain(seed, vectors, Network.mainnet, HdKeyType.z);

        const addressVectors: [string, string, string, string][] = [
            [
                "m/84'/0'/0'/0/0",
                "KyZpNDKnfs94vbrwhJneDi77V6jF64PWPF8x5cdJb8ifgg2DUc9d",
                "0330d54fd0dd420a6e5f8d3624f5f3482cae350f79d5f0753bf5beef9c2d91af3c",
                "bc1qcr8te4kr609gcawutmrza0j4xv80jy8z306fyu",
            ],
            [
                "m/84'/0'/0'/0/1",
                "Kxpf5b8p3qX56DKEe5NqWbNUP9MnqoRFzZwHRtsFqhzuvUJsYZCy",
                "03e775fd51f0dfb8cd865d9ff1cca2a158cf651fe997fdc9fee9c1d3b5e995ea77",
                "bc1qnjg0jd8228aq7egyzacy8cys3knf9xvrerkf9g",
            ],
            [
                "m/84'/0'/0'/1/0",
                "KxuoxufJL5csa1Wieb2kp29VNdn92Us8CoaUG3aGtPtcF3AzeXvF",
                "03025324888e429ab8e3dbaf1f7802648b9cd01e9b418485c5fa4c1b9b5700e1a6",
                "bc1q8c6fshw2dlwun7ekn9qwf37cu2rn755upcp6el",
            ],
        ];

        for (const [path, wif, pubkey, address] of addressVectors) {
            it("address " + path, () => {
                const prvKey = HdPrivateKey.fromPath(path, seed, Network.mainnet, HdKeyType.z);

                expect(prvKey.privateKey.toHex()).to.equal(Wif.decode(wif).privateKey.toString("hex")); // prettier-ignore
                expect(prvKey.toPubKey().publicKey.toHex()).to.equal(pubkey); // prettier-ignore
                expect(prvKey.toPubKey().toAddress()).to.equal(address);
            });
        }
    });
});
