"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Address = void 0;
const bufio_1 = require("@node-lightning/bufio");
const Base58Check_1 = require("./Base58Check");
const Bech32_1 = require("./Bech32");
const BitcoinError_1 = require("./BitcoinError");
const BitcoinErrorCode_1 = require("./BitcoinErrorCode");
const Network_1 = require("./Network");
/**
 * Address encoding/decoding utility for legacy and segwit addresses.
 */
class Address {
    /**
     * Encodes a P2PKH or P2SH address using Base58Check in the format
     * - 1-byte: prefix
     * - 20-byte: hash160 of data
     * - 4-byte: checksum from hash256 of data
     * @param prefix - 1-byte prefix from a Network config (either p2)
     * @param hash
     * @returns
     */
    static encodeBase58(prefix, hash) {
        if (hash.length !== 20) {
            throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.Hash160Invalid, { hash });
        }
        const w = new bufio_1.BufferWriter(Buffer.alloc(21));
        w.writeUInt8(prefix);
        w.writeBytes(hash);
        return Base58Check_1.Base58Check.encode(w.toBuffer());
    }
    /**
     * Decodes a Base58Check encoded legacy address and determines the
     * network. This function returns the hash160 and the prefix.
     * @param encoded base58check encoded string
     * @returns
     */
    static decodeBase58(encoded) {
        const data = Base58Check_1.Base58Check.decode(encoded);
        const prefix = data[0];
        const hash = data.slice(1);
        if (hash.length !== 20) {
            throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.Hash160Invalid, { hash });
        }
        for (const network of Network_1.Network.all) {
            if (network.p2pkhPrefix === prefix || network.p2shPrefix === prefix) {
                return { network, prefix, hash };
            }
        }
        throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.UnknownAddressPrefix, { encoded });
    }
    /**
     * Encodes a native segwit P2WPKH or P2WSH address using bech32.
     * @param prefix hrp part of the address
     * @param version segwit version
     * @param program witness program
     * @returns
     */
    static encodeBech32(prefix, version, program) {
        const words = [version].concat(Bech32_1.Bech32.bufferToWords(program, true));
        return Bech32_1.Bech32.encode(prefix, words, Bech32_1.Bech32Version.Bech32);
    }
    /**
     * Decodes a bech32 address and returns the corresponding network,
     * segwit version, and witness program.
     * @param encoded
     * @returns
     */
    static decodeBech32(encoded) {
        const { hrp, words, version: checksum } = Bech32_1.Bech32.decode(encoded);
        const version = words[0];
        if (version < 0 || version > 16) {
            throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.InvalidSegwitVersion, { encoded, version });
        }
        if (version === 0 && checksum !== Bech32_1.Bech32Version.Bech32) {
            throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.InvalidBech32Encoding);
        }
        if (version > 0 && checksum !== Bech32_1.Bech32Version.Bech32m) {
            throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.InvalidBech32Encoding);
        }
        const program = Bech32_1.Bech32.wordsToBuffer(words.slice(1), false);
        if (version === 0 && program.length !== 20 && program.length !== 32) {
            throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.InvalidWitnessProgram, {
                encoded,
                version,
                program,
                length: program.length,
            });
        }
        if (program.length < 2 || program.length > 40) {
            throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.InvalidWitnessProgram, {
                encoded,
                version,
                program,
                length: program.length,
            });
        }
        for (const network of Network_1.Network.all) {
            if (hrp === network.p2wpkhPrefix)
                return { network, version, program };
        }
        throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.UnknownAddressPrefix, { encoded, hrp });
    }
}
exports.Address = Address;
//# sourceMappingURL=Address.js.map