/// <reference types="node" />
export declare enum Bech32Version {
    Bech32 = 1,
    Bech32m = 734539939
}
/**
 * Encoding utilities for Bech32 encodings as defined in BIP173:
 * https://github.com/bitcoin/bips/blob/master/bip-0173.mediawiki#Bech32
 *
 * This encoding uses a human readable part that contain between 1-83
 * US-ASCII characters with char values in the range [33-126].
 *
 * The encoding includes a separator, which is the last "1" in the encoding.
 *
 * Data must be at least 6-characters and is either lowercase or uppercase
 * and excludes the characters "1", "b", "i", and "o".
 */
export declare class Bech32 {
    /**
     * Encodes the HRP and data according and appends a 6-character
     * checksum.
     * @param hrp human readable part
     * @param words 5-bit words
     * @param version bech32 or bech32m
     * @returns BECH32 encoded string
     */
    static encode(hrp: string, words: number[], version?: Bech32Version): string;
    /**
     * Decodes the BECH32 string into the human readable part and a
     * sequence of 5-bit words. The input can be either upper case or
     * lower case. The last 6-characters are the checksum.
     * @param encoded
     * @returns
     */
    static decode(encoded: string): {
        hrp: string;
        words: number[];
        version: Bech32Version;
    };
    /**
     * Converts 5-bit words into 8-byte words in a Buffer. When padding
     * is enabled incomplete input words are padded to make a complete
     * output byte. For example, if there are 4-words, this makes 20-bits
     * but requires 3-bytes (24-bits) of data. Padding includes this 3rd
     * byte but include zero bits.
     * @param words
     * @param pad
     * @returns
     */
    static wordsToBuffer(words: number[], pad: boolean): Buffer;
    /**
     * Converts a buffer of 8-bit words into 5-bit words. Padding is
     * defaulted to false. For instance if there are 3-bytes (24-bits)
     * we only have 4 5-bit words (20-bits).  The remaining 4-bits will
     * be dropped when pad is false. Otherwise you will get a zero
     * word.
     * @param buffer
     * @param pad default=false
     * @returns
     */
    static bufferToWords(buffer: Buffer, pad: boolean): number[];
    /**
     * Calculates the number of words needed to store the number.
     * In bech32, each word is 5 bits (2^5 = 32). To properly
     * encode a number, we need to determine how many words
     * are necessary to encode the value.
     *
     * @param num integer value
     * @returns number of words
     */
    static sizeofNum(num: number): number;
    /**
     * Calculates the number of words needed to store
     * the supplied number of bits. In bech32, each words is 5 bits
     * (2^5 = 32).
     *
     * @param bits number of bits
     * @returns number of words
     */
    static sizeofBits(bits: number): number;
    /**
     * Calculates the number of words needeed to store
     * the supplied number of bytes. In bech32, each words is 5 bits
     * (2^5 = 32). We first convert the bytes to bits and then
     * convert to words.
     *
     * @param bytes
     * @returns number of words
     */
    static sizeofBytes(bytes: number): number;
}
