/// <reference types="node" />
/**
 * Bits encodes the proof-of-work necessary in this block. It contains
 * two parts: exponent and coefficient. It is a succinct way to express
 * a really large number. It can be converted into a `target` or the
 * `difficulty`.
 */
export declare class Bits {
    readonly coefficient: number;
    readonly exponent: number;
    /**
     * Parses nBits from a LE buffer as represented inside of a block
     * where first three bytes are the coefficient and the last byte
     * is the exponent.
     * @param buf
     * @returns
     */
    static fromBuffer(buf: Buffer): Bits;
    /**
     * Converts a target into a `Bits` instance. This is the inverse of
     * the `target` property on a `Bits` instance. The algorithm to
     * convert a target is to write the number into a Buffer in
     * big-endian. The required bytes is the exponent and the
     * coefficient is the top three most bytes. If the topmost bit is
     * set then we
     * @param target
     */
    static fromTarget(target: bigint): Bits;
    /**
     * Constructs a new `Bits` instance.
     * @param coefficient
     * @param exponent
     */
    constructor(coefficient: number, exponent: number);
    /**
     * Serializes the bits to a Buffer in the format used for RPC block
     * serialization. This means that hte lower three bytes are the
     * coefficient little-endian and the upper byte is the exponent.
     * One final piece of complexity is that if the coefficient is
     * greater than 0x800000 then we pad it and
     */
    toBuffer(): Buffer;
    /**
     * Calculates the proof-of-work requirements from the bits. Target
     * is calculated as:
     *
     * ```
     * target = coefficient * 256^(exponent-3)
     * ```
     *
     * The target looks like:
     * 0000000000000000013ce9000000000000000000000000000000000000000000
     *
     */
    get target(): bigint;
    /**
     * Difficulty is a more readable version of the target and makes the
     * targets more easy to compare and comprehend. The genesis block
     * had a difficulty of 1.
     *
     * It is calculated with the formula:
     *
     * ```
     * difficulty = 0xffff * 256 ^ (0x1d - 3) / target
     * ```
     */
    get difficulty(): bigint;
}
