"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Block = void 0;
const bufio_1 = require("@node-lightning/bufio");
const BlockHeader_1 = require("./BlockHeader");
const Tx_1 = require("./Tx");
/**
 * This class represents a Bitcoin block including fully constructed
 * transactions. It can be used to deserialize a raw transaction receive
 * from a peer or deserialized from HEX.
 *
 * This class extends the `BlockHeader` type since all properties on
 * a `BlockHeader` are part of a `Block`.
 */
class Block extends BlockHeader_1.BlockHeader {
    constructor(version, previousBlockHash, merkleRoot, timestamp, bits, nonce, txs) {
        super(version, previousBlockHash, merkleRoot, timestamp, bits, nonce);
        this.txs = txs;
    }
    /**
     * Parses a Block with full transactions information by first
     * parsing the `BlockHeader` then the transactions.
     * @param buf
     */
    static fromBuffer(buf) {
        const r = new bufio_1.BufferReader(buf);
        // parse the header
        const header = BlockHeader_1.BlockHeader.fromBuffer(r.readBytes(BlockHeader_1.BlockHeader.ByteSize));
        // number of transactions
        const numTxs = Number(r.readVarUint());
        // read txs as a stream
        const txsStream = bufio_1.StreamReader.fromBuffer(r.readBytes());
        const txs = [];
        for (let i = 0; i < numTxs; i++) {
            txs.push(Tx_1.Tx.decode(txsStream));
        }
        // return block
        return new Block(header.version, header.previousBlockHash, header.merkleRoot, header.timestamp, header.bits, header.nonce, txs);
    }
    /**
     * Serializes the `Block` into a Buffer by first serializing the
     * `BlockHeader` followed by a `varint` that encodes the number of
     * transactions followed by a sequence of `Tx`.
     */
    toBuffer() {
        const bw = new bufio_1.BufferWriter();
        // serialize the header
        bw.writeBytes(super.toBuffer());
        // serialize tx count
        bw.writeVarInt(this.txs.length);
        // serialize txs
        for (const tx of this.txs) {
            bw.writeBytes(tx.serialize());
        }
        return bw.toBuffer();
    }
}
exports.Block = Block;
//# sourceMappingURL=Block.js.map