/// <reference types="node" />
import { HdKeyType } from "./HdKeyType";
import { HdPublicKey } from "./HdPublicKey";
import { Network } from "./Network";
import { PrivateKey } from "./PrivateKey";
/**
 * Represnts a hierarchical deterministic extended private key as defined
 * in BIP32. This class contains helper methods to derive child keys and
 * manage derivation.
 */
export declare class HdPrivateKey {
    /**
     * Constructs an extended private key from a derivation path by
     * performing a sequence of private key derivations.
     *
     * For example: `m/0'/1`
     *
     * * Must start at the master key `m`
     * * Hardened indices include a tick, `'`, eg: `0'` or `100'`
     * * Each level is split with a `/`
     *
     * Derivation requires the seed, network, and type to correctly
     * derive the private key.
     *
     * @param path string path
     * @param seed 32-byte seed
     * @param network network that the private key belongs to
     * @param type type of HD key (x, y, z) which defaults to x when a
     * BIP49 or BIP84 path is not detected
     * @returns the extended private key at the defined path
     */
    static fromPath(path: string, seed: Buffer, network: Network, type?: HdKeyType): HdPrivateKey;
    /**
     * Generates the master extended private key from the seed.
     *
     * @remarks
     * The master key has a depth of 0 and an parent fingerprint of
     * 0x00000000. The master key is generated from the HMAC-SHA512
     * using the key "Bitcoin seed" and the 32-byte seed as the data
     * part.
     *
     * @param seed 32-byte seed
     * @param network network the key belongs to
     * @param type HD key type (x, y, z)
     * @returns master extended pivate key
     */
    static fromSeed(seed: Buffer, network: Network, type?: HdKeyType): HdPrivateKey;
    /**
     * Decodes an extended private key from the serialization format
     * defined in BIP32.
     *
     * For example:
     * xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi
     *
     * @remarks
     *
     * The value uses a prefix of xprv, yprv, or zprv and is Base58Check.
     *
     * ```
     * The format includes:
     * [4 byte]: version
     * [1 byte]: depth
     * [4 byte]: parent fingerprint
     * [4 byte]: number
     * [32 byte]: chaincode
     * [33 byte]: key
     * ```
     *
     * The 33-byte key values uses a 0x00 prefix plus the 32-byte private
     * key.
     *
     * @param input encoded input
     *
     * @throws {@link BitcoinError} throws when there is an invalid
     * encoding, bad checksum, or if you attempt to decode a public key.
     *
     * @returns an instance of the extended private key
     */
    static decode(input: string): HdPrivateKey;
    /**
     * The type of extended key. This ensures that xprv, yprv, and
     * zprv values are not used in the same path.
     */
    type: HdKeyType;
    /**
     * The depth of the derivation path. Depth of 0 is the master key.
     * Maximum depth is 255 since this is encoded using a single byte.
     */
    depth: number;
    /**
     * The parents fingerprint which is the first 4-bytes of the hash160
     * of the compressed public key. The fingerprint of the parent only
     * serves as a fast way to detect parent and child keys and software
     * must be willing to deal with collisions.
     */
    parentFingerprint: Buffer;
    /**
     * Specifies the child key number. Values between 0 to 2^31-1 are
     * non-hardened. Values between 2^31 to 2^32-1 are hardened.
     */
    number: number;
    /**
     * The extended private key's chaincode
     */
    chainCode: Buffer;
    /**
     * The underlying private key instance for extended private key
     */
    privateKey: PrivateKey;
    private _fingerprint;
    /**
     * The network the extended private key belongs to
     */
    get network(): Network;
    /**
     * The numeric version of the extended private key. This value is
     * determiend by the network and is used to encode a prefix when
     * the extended private key is encoded as a string.
     */
    get version(): number;
    /**
     * Returns true when the key is a hardened key. A key is hardened
     * when it has an index between 2^31 and 2^32-1
     */
    get isHardened(): boolean;
    /**
     * Lazy loads the fingerprint of the current extended private key
     */
    get fingerprint(): Buffer;
    /**
     * Derives the child private key at the specified index. This
     * method is capable of deriving both hardened and non-hardened
     * child keys.
     *
     * @remarks
     *
     * @param i number of key to derive
     * @throws if this generates an invalid key either because the tweak
     * was invalid or the resulting private key is invalid.
     */
    derive(i: number): HdPrivateKey;
    /**
     * Derives a hardened child key at the specified index. This method
     * simplifies derivation of hardened keys be treating the argument
     * as a hardened index. For example: `.deriveHardened(0)` derives
     * `0'` at index `0x80000000`.
     * @param i index of hardened key
     * @throws if this generates an invalid key either because the tweak
     * was invalid or the resulting private key is invalid.
     */
    deriveHardened(i: number): HdPrivateKey;
    /**
     * Constructs the corresponding {@link HdPublicKey}.
     *
     * The derivation is defined in BIP32 using HMAC-SHA512
     * {@link https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki}
     *
     * The derivation uses the parent's chaincode as the `key`. Then uses
     * for the `data`:
     *
     * * non-hardened: `parent_pub_key || i`
     * * hardened: `0x00 || parent_priv_key || i`
     *
     * As such the data is always 37-bytes.
     *
     * The resulting 64-bytes are split where where the left half is
     * used to tweakAdd against the parent private key. The right half
     * is the child's chain_code.
     *
     * @returns The corresponding extended public key
     */
    toPubKey(): HdPublicKey;
    /**
     * Returns the compressed WIF encoding
     * @returns
     */
    toWif(): string;
    /**
     * Returns the address encoded according to the type of HD key.
     * For x-type this returns a base58 encoded P2PKH address.
     * For y-type this returns a base58 encoded P2SH-P2WPKH address.
     * For z-type this returns a bech32 encoded P2WPKH address.
     */
    toAddress(): string;
    /**
     * Encodes the {@link HdPrivateKey} according to BIP32.
     *
     * For example:
     * xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi
     *
     * @remarks
     * The value uses a prefix of xprv, yprv, or zprv and is Base58Check
     * encoded.
     *
     * The format includes:
     * ```
     * [4 byte]: version
     * [1 byte]: depth
     * [4 byte]: parent fingerprint
     * [4 byte]: number
     * [32 byte]: chaincode
     * [33 byte]: key
     * ```
     *
     * For private keys, the 33-byte key value is prefixed with 0x00.
     * @returns Base58Check encoded extended private key
     */
    encode(): string;
    /**
     * Returns a Buffer of the underlying private key encoded as a big
     * endian integer. This is sugar for `key.privateKey.toBuffer()`.
     */
    toBuffer(): Buffer;
    /**
     * Returns a hex string of the underlying private key. This is sugar
     * for `key.privateKey.toHex()`.
     */
    toHex(): string;
}
