"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.HdPrivateKey = void 0;
const crypto = __importStar(require("@node-lightning/crypto"));
const bufio_1 = require("@node-lightning/bufio");
const BitcoinError_1 = require("./BitcoinError");
const BitcoinErrorCode_1 = require("./BitcoinErrorCode");
const HdKeyCodec_1 = require("./HdKeyCodec");
const HdKeyType_1 = require("./HdKeyType");
const HdPublicKey_1 = require("./HdPublicKey");
const PrivateKey_1 = require("./PrivateKey");
const MAX_INDEX = 0xffffffff;
const HARDENED_INDEX = 0x80000000;
const BIP49_PURPOSE = 0x80000031; // 49'
const BIP84_PURPOSE = 0x80000054; // 84'
/**
 * Represnts a hierarchical deterministic extended private key as defined
 * in BIP32. This class contains helper methods to derive child keys and
 * manage derivation.
 */
class HdPrivateKey {
    /**
     * Constructs an extended private key from a derivation path by
     * performing a sequence of private key derivations.
     *
     * For example: `m/0'/1`
     *
     * * Must start at the master key `m`
     * * Hardened indices include a tick, `'`, eg: `0'` or `100'`
     * * Each level is split with a `/`
     *
     * Derivation requires the seed, network, and type to correctly
     * derive the private key.
     *
     * @param path string path
     * @param seed 32-byte seed
     * @param network network that the private key belongs to
     * @param type type of HD key (x, y, z) which defaults to x when a
     * BIP49 or BIP84 path is not detected
     * @returns the extended private key at the defined path
     */
    static fromPath(path, seed, network, type) {
        const parts = path.split("/");
        // verify that we start with 'm' and that there are at most 255
        // sub-parts
        if (parts[0] !== "m" || parts.length > 255) {
            throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.InvalidHdPath, path);
        }
        // calculate the nums before we do anything else
        const nums = [];
        for (let i = 1; i < parts.length; i++) {
            const part = parts[i];
            // extract number
            const hardened = part.endsWith("'");
            const num = hardened
                ? HARDENED_INDEX + Number(part.substring(0, part.length - 1))
                : Number(part);
            // validate path was correct
            if (isNaN(num) || num < 0 || num > MAX_INDEX || (!hardened && num >= HARDENED_INDEX)) {
                throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.InvalidHdPath, path);
            }
            nums.push(num);
        }
        // attempt to detect the type if one was not supplied...
        if (!type) {
            // purpose is 84', type is z
            if (nums[0] && nums[0] === BIP84_PURPOSE) {
                type = HdKeyType_1.HdKeyType.z;
            }
            // purpose is 49', type is y
            else if (nums[0] && nums[0] === BIP49_PURPOSE) {
                type = HdKeyType_1.HdKeyType.y;
            }
            // otherwise it's x
            else {
                type = HdKeyType_1.HdKeyType.x;
            }
        }
        // generate the master key
        let key = HdPrivateKey.fromSeed(seed, network, type);
        // perform derivations from the master
        for (const num of nums) {
            key = key.derive(num);
        }
        return key;
    }
    /**
     * Generates the master extended private key from the seed.
     *
     * @remarks
     * The master key has a depth of 0 and an parent fingerprint of
     * 0x00000000. The master key is generated from the HMAC-SHA512
     * using the key "Bitcoin seed" and the 32-byte seed as the data
     * part.
     *
     * @param seed 32-byte seed
     * @param network network the key belongs to
     * @param type HD key type (x, y, z)
     * @returns master extended pivate key
     */
    static fromSeed(seed, network, type = HdKeyType_1.HdKeyType.x) {
        const l = crypto.hmac(Buffer.from("Bitcoin seed"), seed, "sha512");
        const key = new HdPrivateKey();
        key.type = type;
        key.depth = 0;
        key.number = 0;
        key.parentFingerprint = Buffer.from([0, 0, 0, 0]);
        key.chainCode = l.slice(32);
        key.privateKey = new PrivateKey_1.PrivateKey(l.slice(0, 32), network);
        return key;
    }
    /**
     * Decodes an extended private key from the serialization format
     * defined in BIP32.
     *
     * For example:
     * xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi
     *
     * @remarks
     *
     * The value uses a prefix of xprv, yprv, or zprv and is Base58Check.
     *
     * ```
     * The format includes:
     * [4 byte]: version
     * [1 byte]: depth
     * [4 byte]: parent fingerprint
     * [4 byte]: number
     * [32 byte]: chaincode
     * [33 byte]: key
     * ```
     *
     * The 33-byte key values uses a 0x00 prefix plus the 32-byte private
     * key.
     *
     * @param input encoded input
     *
     * @throws {@link BitcoinError} throws when there is an invalid
     * encoding, bad checksum, or if you attempt to decode a public key.
     *
     * @returns an instance of the extended private key
     */
    static decode(input) {
        const result = HdKeyCodec_1.HdKeyCodec.decode(input);
        if (!(result instanceof HdPrivateKey)) {
            throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.InvalidHdPrivateKey, input);
        }
        else {
            return result;
        }
    }
    /**
     * The network the extended private key belongs to
     */
    get network() {
        return this.privateKey.network;
    }
    /**
     * The numeric version of the extended private key. This value is
     * determiend by the network and is used to encode a prefix when
     * the extended private key is encoded as a string.
     */
    get version() {
        switch (this.type) {
            case HdKeyType_1.HdKeyType.x:
                return this.network.xprvVersion;
            case HdKeyType_1.HdKeyType.y:
                return this.network.yprvVersion;
            case HdKeyType_1.HdKeyType.z:
                return this.network.zprvVersion;
        }
    }
    /**
     * Returns true when the key is a hardened key. A key is hardened
     * when it has an index between 2^31 and 2^32-1
     */
    get isHardened() {
        return this.number >= HARDENED_INDEX;
    }
    /**
     * Lazy loads the fingerprint of the current extended private key
     */
    get fingerprint() {
        if (!this._fingerprint) {
            this._fingerprint = crypto.hash160(this.privateKey.toPubKey(true).toBuffer());
        }
        return this._fingerprint;
    }
    /**
     * Derives the child private key at the specified index. This
     * method is capable of deriving both hardened and non-hardened
     * child keys.
     *
     * @remarks
     *
     * @param i number of key to derive
     * @throws if this generates an invalid key either because the tweak
     * was invalid or the resulting private key is invalid.
     */
    derive(i) {
        const result = new HdPrivateKey();
        result.type = this.type;
        result.depth = this.depth + 1;
        result.number = i;
        result.parentFingerprint = Buffer.from(this.fingerprint.slice(0, 4));
        const data = new bufio_1.BufferWriter(Buffer.alloc(37));
        // hardened
        if (i >= HARDENED_INDEX) {
            data.writeUInt8(0);
            data.writeBytes(this.privateKey.toBuffer());
            data.writeUInt32BE(i);
        }
        // normal child
        else {
            data.writeBytes(this.privateKey.toPubKey(true).toBuffer());
            data.writeUInt32BE(i);
        }
        const l = crypto.hmac(this.chainCode, data.toBuffer(), "sha512");
        const ll = l.slice(0, 32);
        const lr = l.slice(32);
        result.privateKey = this.privateKey.tweakAdd(ll);
        result.chainCode = lr;
        return result;
    }
    /**
     * Derives a hardened child key at the specified index. This method
     * simplifies derivation of hardened keys be treating the argument
     * as a hardened index. For example: `.deriveHardened(0)` derives
     * `0'` at index `0x80000000`.
     * @param i index of hardened key
     * @throws if this generates an invalid key either because the tweak
     * was invalid or the resulting private key is invalid.
     */
    deriveHardened(i) {
        if (i < HARDENED_INDEX) {
            i = i + HARDENED_INDEX;
        }
        return this.derive(i);
    }
    /**
     * Constructs the corresponding {@link HdPublicKey}.
     *
     * The derivation is defined in BIP32 using HMAC-SHA512
     * {@link https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki}
     *
     * The derivation uses the parent's chaincode as the `key`. Then uses
     * for the `data`:
     *
     * * non-hardened: `parent_pub_key || i`
     * * hardened: `0x00 || parent_priv_key || i`
     *
     * As such the data is always 37-bytes.
     *
     * The resulting 64-bytes are split where where the left half is
     * used to tweakAdd against the parent private key. The right half
     * is the child's chain_code.
     *
     * @returns The corresponding extended public key
     */
    toPubKey() {
        const result = new HdPublicKey_1.HdPublicKey();
        result.type = this.type;
        result.depth = this.depth;
        result.number = this.number;
        result.parentFingerprint = Buffer.from(this.parentFingerprint);
        result.chainCode = Buffer.from(this.chainCode);
        result.publicKey = this.privateKey.toPubKey(true);
        return result;
    }
    /**
     * Returns the compressed WIF encoding
     * @returns
     */
    toWif() {
        return this.privateKey.toWif(true);
    }
    /**
     * Returns the address encoded according to the type of HD key.
     * For x-type this returns a base58 encoded P2PKH address.
     * For y-type this returns a base58 encoded P2SH-P2WPKH address.
     * For z-type this returns a bech32 encoded P2WPKH address.
     */
    toAddress() {
        return this.toPubKey().toAddress();
    }
    /**
     * Encodes the {@link HdPrivateKey} according to BIP32.
     *
     * For example:
     * xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi
     *
     * @remarks
     * The value uses a prefix of xprv, yprv, or zprv and is Base58Check
     * encoded.
     *
     * The format includes:
     * ```
     * [4 byte]: version
     * [1 byte]: depth
     * [4 byte]: parent fingerprint
     * [4 byte]: number
     * [32 byte]: chaincode
     * [33 byte]: key
     * ```
     *
     * For private keys, the 33-byte key value is prefixed with 0x00.
     * @returns Base58Check encoded extended private key
     */
    encode() {
        return HdKeyCodec_1.HdKeyCodec.encode(this);
    }
    /**
     * Returns a Buffer of the underlying private key encoded as a big
     * endian integer. This is sugar for `key.privateKey.toBuffer()`.
     */
    toBuffer() {
        return this.privateKey.toBuffer();
    }
    /**
     * Returns a hex string of the underlying private key. This is sugar
     * for `key.privateKey.toHex()`.
     */
    toHex() {
        return this.privateKey.toHex();
    }
}
exports.HdPrivateKey = HdPrivateKey;
//# sourceMappingURL=HdPrivateKey.js.map