/// <reference types="node" />
import { Network } from "./Network";
import { PublicKey } from "./PublicKey";
/**
 * This class encapsulates a 32-byte buffer containing a big-endian
 * encoded number between 0x1 and
 * 0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364140
 * that is a private key for the secp256k1 elliptic curve.
 */
export declare class PrivateKey {
    readonly network: Network;
    /**
     * Decodes a WIF encoded string into the private key as well as the
     * public key. The inclusion of the public key allows us to capture
     * if the public key was compressed.
     * @param input WIF encoded string
     * @returns A tuple of the private key and public key
     * @throw {@link BitcoinError} Throws for invalid network prefix or
     * invalid private or public keys.
     */
    static fromWif(input: string): [PrivateKey, PublicKey];
    private _buffer;
    /**
     * Constructs a new private key instance and requires a valid private
     * key value.
     * @param buffer 32-byte buffer containing a valid private key
     * @param network The network that the private key belongs to
     */
    constructor(buffer: Buffer, network: Network);
    /**
     * Converts the instance into a big-endian encoded 32-byte buffer
     * @returns 32-byte buffer
     */
    toBuffer(): Buffer;
    /**
     * Converts the instance into a big-endian encoded hex string
     */
    toHex(): string;
    /**
     * Converts the private key into the corresponding public key
     */
    toPubKey(compressed: boolean): PublicKey;
    /**
     * Converts the private key to WIF
     */
    toWif(compressed: boolean): string;
    /**
     * Tweaks a private key by adding a value to it.
     * The equation is: e + t.
     *
     * @param tweak a 32-byte tweak
     * @returns new instance of PrivateKey with the tweaked value
     */
    tweakAdd(tweak: Buffer): PrivateKey;
    /**
     * Tweaks a private key by multiplying it. The equation is: e * t.
     *
     * @param tweak a 32-byte tweak
     * @returns new instance of PrivateKey with the tweaked value
     */
    tweakMul(tweak: Buffer): PrivateKey;
}
