"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PrivateKey = void 0;
const crypto = __importStar(require("@node-lightning/crypto"));
const _1 = require(".");
const BitcoinError_1 = require("./BitcoinError");
const BitcoinErrorCode_1 = require("./BitcoinErrorCode");
const PublicKey_1 = require("./PublicKey");
/**
 * This class encapsulates a 32-byte buffer containing a big-endian
 * encoded number between 0x1 and
 * 0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364140
 * that is a private key for the secp256k1 elliptic curve.
 */
class PrivateKey {
    /**
     * Constructs a new private key instance and requires a valid private
     * key value.
     * @param buffer 32-byte buffer containing a valid private key
     * @param network The network that the private key belongs to
     */
    constructor(buffer, network) {
        this.network = network;
        if (!crypto.validPrivateKey(buffer)) {
            throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.InvalidPrivateKey, { key: buffer });
        }
        this._buffer = buffer;
    }
    /**
     * Decodes a WIF encoded string into the private key as well as the
     * public key. The inclusion of the public key allows us to capture
     * if the public key was compressed.
     * @param input WIF encoded string
     * @returns A tuple of the private key and public key
     * @throw {@link BitcoinError} Throws for invalid network prefix or
     * invalid private or public keys.
     */
    static fromWif(input) {
        const decoded = _1.Wif.decode(input);
        const network = _1.Wif.decodePrefix(decoded.prefix);
        const prvkey = new PrivateKey(decoded.privateKey, network);
        const pubkey = prvkey.toPubKey(decoded.compressed);
        return [prvkey, pubkey];
    }
    /**
     * Converts the instance into a big-endian encoded 32-byte buffer
     * @returns 32-byte buffer
     */
    toBuffer() {
        return this._buffer;
    }
    /**
     * Converts the instance into a big-endian encoded hex string
     */
    toHex() {
        return this._buffer.toString("hex");
    }
    /**
     * Converts the private key into the corresponding public key
     */
    toPubKey(compressed) {
        const result = crypto.getPublicKey(this._buffer, compressed);
        return new PublicKey_1.PublicKey(result, this.network);
    }
    /**
     * Converts the private key to WIF
     */
    toWif(compressed) {
        return _1.Wif.encode(this.network.wifPrefix, this._buffer, compressed);
    }
    /**
     * Tweaks a private key by adding a value to it.
     * The equation is: e + t.
     *
     * @param tweak a 32-byte tweak
     * @returns new instance of PrivateKey with the tweaked value
     */
    tweakAdd(tweak) {
        const result = crypto.privateKeyTweakAdd(this._buffer, tweak);
        return new PrivateKey(result, this.network);
    }
    /**
     * Tweaks a private key by multiplying it. The equation is: e * t.
     *
     * @param tweak a 32-byte tweak
     * @returns new instance of PrivateKey with the tweaked value
     */
    tweakMul(tweak) {
        const result = crypto.privateKeyTweakMul(this._buffer, tweak);
        return new PrivateKey(result, this.network);
    }
}
exports.PrivateKey = PrivateKey;
//# sourceMappingURL=PrivateKey.js.map