/// <reference types="node" />
import { Network } from "./Network";
/**
 * This class represents a point on an secp256k1 elliptic curve and acts
 * as a public key along with the network for which the key belongs. The
 * public key can be serialized into a buffer using the SEC compressed
 * and uncompressed format.
 */
export declare class PublicKey {
    readonly network: Network;
    /**
     * Returns the public key associated with the WIF input
     * @param input WIF encoded private key
     * @returns public key assocated with the WIF encoding
     */
    static fromWif(input: string): PublicKey;
    readonly compressed: boolean;
    private readonly _buffer;
    /**
     * Constructs a new instance of a PublicKey and requires a valid key
     * or an exception will be throw.
     * @param buffer A valid 33-byte or 65-byte SEC encoded public key
     * @param network The corresponding network where the public key belongs
     */
    constructor(buffer: Buffer, network: Network);
    /**
     * Tweaks a public key by adding tweak * G to the point. The equation is
     * T = P + t*G
     *
     * @param tweak 32-byte scalar value that is multiplied by G
     * @returns a new instance of PublicKey containing the tweaked value
     */
    tweakAdd(tweak: Buffer): PublicKey;
    /**
     * Tweaks a public key by multiplying it against a scalar. The equation is
     * T = P * t
     *
     * @param tweak 32-byte tweak to multiply against the public key
     * @returns a new instance of PublicKey containing the tweaked value
     */
    tweakMul(tweak: Buffer): PublicKey;
    /**
     * Adds two public key points together.
     * @param other The other public key that should be added
     * @returns a new instance of PublicKey containing the added points
     */
    add(other: PublicKey): PublicKey;
    /**
     * Returns the hash160 of the public key
     * @returns 20-byte hash160 of the public key
     */
    hash160(): Buffer;
    /**
     * Convert the public key to compressed or uncompressed.
     * @param compressed
     */
    toPubKey(compressed: boolean): PublicKey;
    /**
     * Serializes the PublicKey instance into a 33-byte or 65-byte SEC
     * encoded buffer depending on whether the public key is compressed
     * or uncompressed.
     * @returns 33-byte or 65-byte buffer
     */
    toBuffer(): Buffer;
    /**
     * Serializes the PublicKey instance nito a 33-byte or 65-byte SEC
     * encoded hex value.
     * @returns
     */
    toHex(): string;
    /**
     * Returns a P2PKH for the public key.
     * @returns Base58 encoded Bitcoin address
     */
    toP2pkhAddress(): string;
    /**
     * Returns a nested segwit addrress (P2SH-P2WPKH).
     * @returns Base58 encoded Bitcoin address
     */
    toP2nwpkhAddress(): string;
    /**
     * Returns a P2WPKH address for the public key.
     * @returns bech32 encoded segwit address
     */
    toP2wpkhAddress(): string;
}
