"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PublicKey = void 0;
const crypto = __importStar(require("@node-lightning/crypto"));
const PrivateKey_1 = require("./PrivateKey");
const Script_1 = require("./Script");
const Address_1 = require("./Address");
const BitcoinError_1 = require("./BitcoinError");
const BitcoinErrorCode_1 = require("./BitcoinErrorCode");
/**
 * This class represents a point on an secp256k1 elliptic curve and acts
 * as a public key along with the network for which the key belongs. The
 * public key can be serialized into a buffer using the SEC compressed
 * and uncompressed format.
 */
class PublicKey {
    /**
     * Constructs a new instance of a PublicKey and requires a valid key
     * or an exception will be throw.
     * @param buffer A valid 33-byte or 65-byte SEC encoded public key
     * @param network The corresponding network where the public key belongs
     */
    constructor(buffer, network) {
        this.network = network;
        if (!crypto.validPublicKey(buffer)) {
            throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.PubKeyInvalid, { key: buffer });
        }
        this.compressed = buffer.length === 33;
        this._buffer = buffer;
    }
    /**
     * Returns the public key associated with the WIF input
     * @param input WIF encoded private key
     * @returns public key assocated with the WIF encoding
     */
    static fromWif(input) {
        const [, pubkey] = PrivateKey_1.PrivateKey.fromWif(input);
        return pubkey;
    }
    /**
     * Tweaks a public key by adding tweak * G to the point. The equation is
     * T = P + t*G
     *
     * @param tweak 32-byte scalar value that is multiplied by G
     * @returns a new instance of PublicKey containing the tweaked value
     */
    tweakAdd(tweak) {
        const result = crypto.publicKeyTweakAdd(this._buffer, tweak, this.compressed);
        return new PublicKey(result, this.network);
    }
    /**
     * Tweaks a public key by multiplying it against a scalar. The equation is
     * T = P * t
     *
     * @param tweak 32-byte tweak to multiply against the public key
     * @returns a new instance of PublicKey containing the tweaked value
     */
    tweakMul(tweak) {
        const result = crypto.publicKeyTweakMul(this._buffer, tweak, true);
        return new PublicKey(result, this.network);
    }
    /**
     * Adds two public key points together.
     * @param other The other public key that should be added
     * @returns a new instance of PublicKey containing the added points
     */
    add(other) {
        if (this.network !== other.network) {
            throw new BitcoinError_1.BitcoinError(BitcoinErrorCode_1.BitcoinErrorCode.NetworkMismatch, { me: this, other });
        }
        const result = crypto.publicKeyCombine([this.toBuffer(), other.toBuffer()], this.compressed);
        return new PublicKey(result, this.network);
    }
    /**
     * Returns the hash160 of the public key
     * @returns 20-byte hash160 of the public key
     */
    hash160() {
        return crypto.hash160(this.toBuffer());
    }
    /**
     * Convert the public key to compressed or uncompressed.
     * @param compressed
     */
    toPubKey(compressed) {
        const buffer = crypto.convertPublicKey(this._buffer, compressed);
        return new PublicKey(buffer, this.network);
    }
    /**
     * Serializes the PublicKey instance into a 33-byte or 65-byte SEC
     * encoded buffer depending on whether the public key is compressed
     * or uncompressed.
     * @returns 33-byte or 65-byte buffer
     */
    toBuffer() {
        return Buffer.from(this._buffer);
    }
    /**
     * Serializes the PublicKey instance nito a 33-byte or 65-byte SEC
     * encoded hex value.
     * @returns
     */
    toHex() {
        return this.toBuffer().toString("hex");
    }
    /**
     * Returns a P2PKH for the public key.
     * @returns Base58 encoded Bitcoin address
     */
    toP2pkhAddress() {
        return Address_1.Address.encodeBase58(this.network.p2pkhPrefix, this.hash160());
    }
    /**
     * Returns a nested segwit addrress (P2SH-P2WPKH).
     * @returns Base58 encoded Bitcoin address
     */
    toP2nwpkhAddress() {
        return Script_1.Script.p2wpkhLock(this.hash160()).toP2shAddress(this.network);
    }
    /**
     * Returns a P2WPKH address for the public key.
     * @returns bech32 encoded segwit address
     */
    toP2wpkhAddress() {
        return Address_1.Address.encodeBech32(this.network.p2wpkhPrefix, 0, this.hash160());
    }
}
exports.PublicKey = PublicKey;
//# sourceMappingURL=PublicKey.js.map