"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Value = void 0;
const _1 = require(".");
/**
 * Represents bitcoin value that can be converted to or from multiple
 * formats.
 */
class Value {
    constructor(picoSats) {
        this._picoSats = picoSats;
    }
    /**
     * Creates a value object from value in bitcoin, eg: 1.12345678
     * @param num
     */
    static fromBitcoin(num) {
        return Value.fromSats(Math.trunc(num * 1e8));
    }
    /**
     * Creates a value instance from value in satoshis where 1 satoshis
     * equates to 0.00000001 bitcoin.
     * @param num
     */
    static fromSats(num) {
        return new Value(BigInt(num) * BigInt(1e12));
    }
    /**
     * Creates a value instance from value in millisatoshis, 1/1000 of a
     * satoshi.
     * eg: 123 millisatoshis equates to 0.123 satoshis
     * eg: 123 millisatoshis equates to 0.00000000123 bitcoin
     * @param num
     */
    static fromMilliSats(num) {
        return new Value(BigInt(num) * BigInt(1e9));
    }
    /**
     * Creates a value instance from value in microsatoshis, 1/1e6 of a
     * satoshi.
     * eg: 123 microsatoshis equates to 0.000123 satoshis
     * eg: 123 microsatoshis equates to 0.00000000000123 bitcoin
     * @param num
     */
    static fromMicroSats(num) {
        return new Value(BigInt(num) * BigInt(1e6));
    }
    /**
     * Creates a value instance from value in picosatoshis, 1/1e12 of a
     * satoshi.
     * eg: 123 picosatoshis equates to 0.000000000123 satoshis
     * eg: 123 picosatoshis equates to 0.00000000000000000123 bitcoin
     * @param num
     */
    static fromPicoSats(num) {
        return new Value(BigInt(num));
    }
    /**
     * Generates a value instance of zero
     */
    static zero() {
        return new Value(BigInt(0));
    }
    /**
     * Gets the value in picosatoshis (1/1e12 satoshis)
     */
    get psats() {
        return this._picoSats;
    }
    /**
     * Gets the value in millionth of satoshis (1/1e6 satoshis)
     */
    get microsats() {
        return this._picoSats / BigInt(1e6);
    }
    /**
     * Gets the value in millisatoshis (1/1000 satoshis)
     */
    get msats() {
        return this._picoSats / BigInt(1e9);
    }
    /**
     * Gets the value in satoshis (1/1e8 bitcoin)
     */
    get sats() {
        return this._picoSats / BigInt(1e12);
    }
    /**
     * Gets the value in bitcoin
     */
    get bitcoin() {
        return Math.max(0, Number(this.sats) / 1e8);
    }
    /**
     * Clone via deep copy
     */
    clone() {
        return new Value(this._picoSats);
    }
    /**
     * Returns the value as a string of Bitcoin. This is a fixed eight
     * decimal string, as such value below satoshis will be truncated.
     */
    toString() {
        return this.bitcoin.toFixed(8);
    }
    /**
     * Returns true if the current value is equal to the other value
     * @param other
     * @returns
     */
    eq(other) {
        return other._picoSats === this._picoSats;
    }
    /**
     * Returns true if the current value is not equal to the other value
     * @param other
     * @returns
     */
    neq(other) {
        return other._picoSats !== this._picoSats;
    }
    /**
     * Returns true if the current value is greater than the other value.
     * @param other
     * @returns
     */
    gt(other) {
        return this._picoSats > other._picoSats;
    }
    /**
     * Returns true if the current value is greater than or equal to the
     * other value.
     * @param other
     * @returns
     */
    gte(other) {
        return this._picoSats >= other._picoSats;
    }
    /**
     * Returns true if the current value is less than the other value.
     * @param other
     */
    lt(other) {
        return this._picoSats < other._picoSats;
    }
    /**
     * Returns true if the current value is less than or equal to the
     * other value.
     * @param other
     * @returns
     */
    lte(other) {
        return this._picoSats <= other._picoSats;
    }
    /**
     * Modifies the current instance by adding the other value to the
     * existing value.
     * @param other
     */
    add(other) {
        this._picoSats += other._picoSats;
        return this;
    }
    /**
     * Modifies the current instance by subtracting the other value
     * from our existing value. Since Value is unsigned, this throws
     * if subtraction results in a value that is less than zero.
     * @param other
     */
    sub(other) {
        if (this._picoSats - other._picoSats < 0) {
            throw new _1.BitcoinError(_1.BitcoinErrorCode.ValueUnderflow);
        }
        this._picoSats -= other._picoSats;
        return this;
    }
}
exports.Value = Value;
//# sourceMappingURL=Value.js.map